<?php

declare(strict_types=1);

namespace GuzzleHttp\Exception;

/**
 * Exception thrown when a response transfer fails after response headers are
 * received.
 */
class ResponseTransferException extends ResponseException
{
}
