/*
	Copyright (c) 2004-2016, The JS Foundation All Rights Reserved.
	Available via Academic Free License >= 2.1 OR the modified BSD license.
	see: http://dojotoolkit.org/license for details
*/

//>>built
define("dojo/cldr/nls/ru/number",{"scientificFormat":"#E0","infinity":"∞","superscriptingExponent":"×","list":";","percentSign":"%","minusSign":"-","decimalFormat-short":"000 трлн","nan":"не число","plusSign":"+","currencyFormat":"#,##0.00 ¤","perMille":"‰","approximatelySign":"≈","group":" ","percentFormat":"#,##0 %","decimalFormat-long":"000 триллиона","decimalFormat":"#,##0.###","currencyFormat-short":"000 трлн ¤","timeSeparator":":","decimal":",","exponential":"E"});