@import "variables";

// List Icons
// -------------------------

.#{$md-prefix}-ul {
  padding-left: 0;
  margin-left: $md-li-width;
  list-style-type: none;
  > li { position: relative; }
}
.#{$md-prefix}-li {
  position: absolute;
  left: -$md-li-width;
  width: $md-li-width;
  top: (2em / 14);
  text-align: center;
  &.#{$md-prefix}-lg {
    left: -$md-li-width + (4em / 14);
  }
}
